@@ -1,10 +1,17 @@
-# Copyright (c) 2003-2006 CORE Security Technologies
+################################################################################
+# DEPRECATION WARNING!                                                         #
+# This library will be deprecated soon. You should use impacket.dcerpc.v5      #
+# classes instead                                                              #
+################################################################################
+# Copyright (c) 2003-2012 CORE Security Technologies
 #
 # This software is provided under under a slightly modified version
 # of the Apache Software License. See the accompanying LICENSE file
 # for more information.
 #
-# $Id: samr.py,v 1.7 2006/05/23 21:19:26 gera Exp $
+# $Id$
+#
+# Author: Alberto Solino
 #
 # Description:
 #   SAMR (Security Account Manager Remote) interface implementation.
@@ -15,26 +22,117 @@ from time import strftime, gmtime
 from struct import *
 
 from impacket import ImpactPacket
-import dcerpc
-import ndrutils
+from impacket.dcerpc import ndrutils, dcerpc
+from impacket.structure import Structure
+from impacket.uuid import uuidtup_to_bin
 
-MSRPC_UUID_SAMR   = '\x78\x57\x34\x12\x34\x12\xcd\xab\xef\x00\x01\x23\x45\x67\x89\xac\x01\x00\x00\x00'
+MSRPC_UUID_SAMR   = uuidtup_to_bin(('12345778-1234-ABCD-EF00-0123456789AC', '1.0'))
+
+KNOWN_SIDS = {
+}
 
 OP_NUM_CREATE_USER_IN_DOMAIN    = 0xC
 OP_NUM_ENUM_USERS_IN_DOMAIN     = 0xD
 OP_NUM_CREATE_ALIAS_IN_DOMAIN   = 0xE
 
 def display_time(filetime_high, filetime_low, minutes_utc=0):
+    if filetime_low == 4294967295L:
+        r = "Infinity"
+        return r 
     d = filetime_high*4.0*1.0*(1<<30)
     d += filetime_low
     d *= 1.0e-7
     d -= (369.0*365.25*24*60*60-(3.0*24*60*60+6.0*60*60))
+    if d < 1:
+        r = "Undefined"
+        return r
+    
+    try:
+        gmtime(d)
+    except Exception:
+        d = 0
+
     if minutes_utc == 0:
         r = (strftime("%a, %d %b %Y %H:%M:%S",gmtime(d)), minutes_utc/60)[0]
     else:
         r = "%s GMT %d " % (strftime("%a, %d %b %Y %H:%M:%S",gmtime(d)), minutes_utc/60)
     return r
 
+class MSRPCArray:
+    def __init__(self, id=0, len=0, size=0):
+        self._length = len
+        self._size = size
+        self._id = id
+        self._max_len = 0
+        self._offset = 0
+        self._length2 = 0
+        self._name = ''
+
+    def set_max_len(self, n):
+        self._max_len = n
+    def set_offset(self, n):
+        self._offset = n
+    def set_length2(self, n):
+        self._length2 = n
+    def get_size(self):
+        return self._size
+    def set_name(self, n):
+        self._name = n
+    def get_name(self):
+        return self._name
+    def get_id(self):
+        return self._id
+    def rawData(self):
+        return pack('<HHLLLL', self._length, self._size, 0x12345678, self._max_len, self._offset, self._length2) + self._name.encode('utf-16le')
+
+class MSRPCNameArray:
+    def __init__(self, data = None):
+        self._count = 0
+        self._max_count = 0
+        self._elements = []
+
+        if data: self.load(data)
+
+    def load(self, data):
+        ptr = unpack('<L', data[:4])[0]
+        index = 4
+        if 0 == ptr: # No data. May be a bug in certain versions of Samba.
+            return
+
+        self._count, _, self._max_count = unpack('<LLL', data[index:index+12])
+        index += 12
+
+        # Read each object's header.
+        for i in range(0, self._count):
+            aindex, length, size, _ = unpack('<LHHL', data[index:index+12])
+            self._elements.append(MSRPCArray(aindex, length, size))
+            index += 12
+
+        # Read the objects themselves.
+        for element in self._elements:
+            max_len, offset, curlen = unpack('<LLL', data[index:index+12])
+            index += 12
+            element.set_name(unicode(data[index:index+2*curlen], 'utf-16le'))
+            element.set_max_len(max_len)
+            element.set_offset(offset)
+            element.set_length2(curlen)
+            index += 2*curlen
+            if curlen & 0x1: index += 2 # Skip padding.
+
+    def elements(self):
+        return self._elements
+
+    def rawData(self):
+        ret = pack('<LLLL', 0x74747474, self._count, 0x47474747, self._max_count)
+        pos_ret = []
+        for i in xrange(0, self._count):
+            ret += pack('<L', self._elements[i].get_id())
+            data = self._elements[i].rawData()
+            ret += data[:8]
+            pos_ret += data[8:]
+
+        return ret + pos_ret
+
 class MSRPCUserInfo:
     ITEMS = {'Account Name':0,
              'Full Name':1,
@@ -81,7 +179,7 @@ class MSRPCUserInfo:
         index += 48
         for i in range(0,len(MSRPCUserInfo.ITEMS)-1):
             length, size, id = unpack('<HHL',data[index:index+8])
-            self._items.append(dcerpc.MSRPCArray(length, size, id))
+            self._items.append(MSRPCArray(length, size, id))
             index += 8
 
         index += 24     # salteo los unknowns
@@ -90,7 +188,7 @@ class MSRPCUserInfo:
         self._rid, self._group, self._acct_ctr,_ = unpack('<LLLL',data[index: index+16])
         index += 16
         logon_divisions, _, id = unpack('<HHL',data[index:index+8])
-        self._items.append(dcerpc.MSRPCArray(logon_divisions, _, id))
+        self._items.append(MSRPCArray(logon_divisions, _, id))
         index += 8
         self._bad_pwd_count, self._logon_count, self._country, self._codepage = unpack('<HHHH', data[index: index + 8])
         index += 8
@@ -141,10 +239,7 @@ class MSRPCUserInfo:
     def get_logon_count(self):
         return self._logon_count
     def get_pwd_must_change(self):
-        if self._pwd_must_change_low == 4294967295L:
-            return "Infinity" 
-        else:
-            return display_time(self._pwd_must_change_high, self._pwd_must_change_low)
+        return display_time(self._pwd_must_change_high, self._pwd_must_change_low)
     def is_enabled(self):
         return not (self._acct_ctr & 0x01)
 
@@ -157,15 +252,90 @@ class MSRPCUserInfo:
         print "Group id: %d" % self._group
         print "Bad pwd count: %d" % self._bad_pwd_count
         print "Logon count: %d" % self._logon_count
-        if self._pwd_must_change_low == 4294967295L:
-            print "PWD Must Change: Infinity" 
-        else:
-            print "PWD Must Change: " + display_time(self._pwd_must_change_high, self._pwd_must_change_low)
+        print "PWD Must Change: " + display_time(self._pwd_must_change_high, self._pwd_must_change_low)
         for i in MSRPCUserInfo.ITEMS.keys():
             print i + ': ' + self._items[MSRPCUserInfo.ITEMS[i]].get_name()
         print
         return
 
+class SAMR_RPC_SID_IDENTIFIER_AUTHORITY(Structure):
+    structure = (
+        ('Value','6s'),
+    )
+
+class SAMR_RPC_SID(Structure):
+    structure = (
+        ('Revision','<B'),
+        ('SubAuthorityCount','<B'),
+        ('IdentifierAuthority',':',SAMR_RPC_SID_IDENTIFIER_AUTHORITY),
+        ('SubLen','_-SubAuthority','self["SubAuthorityCount"]*4'),
+        ('SubAuthority',':'),
+    )
+
+    def fromCanonical(self, canonical):
+       items = canonical.split('-')
+       self['Revision'] = int(items[1])
+       self['IdentifierAuthority'] = SAMR_RPC_SID_IDENTIFIER_AUTHORITY()
+       self['IdentifierAuthority']['Value'] = '\x00\x00\x00\x00\x00' + pack('B',int(items[2]))
+       self['SubAuthorityCount'] = len(items) - 3
+       ans = ''
+       for i in range(self['SubAuthorityCount']):
+           ans += pack('<L', int(items[i+3]))
+       self['SubAuthority'] = ans
+
+    def formatCanonical(self):
+       ans = 'S-%d-%d' % (self['Revision'], ord(self['IdentifierAuthority']['Value'][5]))
+       for i in range(self['SubAuthorityCount']):
+           ans += '-%d' % ( unpack('<L',self['SubAuthority'][i*4:i*4+4])[0])
+       return ans
+
+class SAMROpenAlias(Structure):
+    opnum = 27
+    alignment = 4
+    structure = (
+        ('ContextHandle','20s'),
+        ('AccessMask','<L'),
+        ('AliasId','<L'),
+    )
+
+class SAMROpenAliasResponse(Structure):
+    structure = (
+        ('ContextHandle','20s'),
+        ('ErrorCode','<L'),
+    )
+
+class SAMRGetMembersInAlias(Structure):
+    opnum = 33
+    alignment = 4
+    structure = (
+        ('ContextHandle','20s'),
+    )
+
+class SAMRGetMembersInAliasResponse(Structure):
+    structure = (
+        ('BuffSize','_-pEnumerationBuffer','len(self.rawData)-8'),
+        ('Count','<L'),
+        ('pEnumerationBuffer',':'),
+        ('ErrorCode','<L'),
+    )
+
+class SAMREnumerateAliasesInDomain(Structure):
+    opnum = 15
+    alignment = 4
+    structure = (
+        ('ContextHandle','20s'),
+        ('ResumeHandle','<L=0'),
+        ('PreferedMaximumLength','<L=0xffffffff'),
+    )
+
+class SAMREnumerateAliasesInDomainResponse(Structure):
+    structure = (
+        ('ResumeHandle','<L=0'),
+        ('BuffSize','_-pEnumerationBuffer','len(self.rawData)-12'),
+        ('pEnumerationBuffer',':'),
+        ('CountReturned','<L'),
+        ('ErrorCode','<L'),
+    )
 
 class SAMRConnectHeader(ImpactPacket.Header):
     OP_NUM = 0x39
@@ -273,9 +443,9 @@ class SAMRRespEnumDomainHeader(ImpactPacket.Header):
         self.set_long(0, handle, '<')
 
     def get_domains(self):
-        return dcerpc.MSRPCNameArray(self.get_bytes()[4:-8].tostring())
+        return MSRPCNameArray(self.get_bytes()[4:-8].tostring())
     def set_domains(self, domains):
-        assert isinstance(domains, dcerpc.MSRPCNameArray)
+        assert isinstance(domains, MSRPCNameArray)
         self.get_bytes()[4:-8] = array.array('B', domains.rawData())
 
     def get_entries_num(self):
@@ -311,9 +481,9 @@ class SAMRLookupDomainHeader(ImpactPacket.Header):
         self.get_bytes()[:20] = array.array('B', handle)
 
     def get_domain(self):
-        return dcerpc.MSRPCArray(self.get_bytes().tolist()[20:])
+        return MSRPCArray(self.get_bytes().tolist()[20:])
     def set_domain(self, domain):
-        assert isinstance(domain, dcerpc.MSRPCArray)
+        assert isinstance(domain, MSRPCArray)
         self.get_bytes()[20:] = array.array('B', domain.rawData())
 
 
@@ -477,9 +647,9 @@ class SAMRRespEnumDomainUsersHeader(ImpactPacket.Header):
         self.set_long(0, handle, '<')
 
     def get_users(self):
-        return dcerpc.MSRPCNameArray(self.get_bytes()[4:-8].tostring())
+        return MSRPCNameArray(self.get_bytes()[4:-8].tostring())
     def set_users(self, users):
-        assert isinstance(users, dcerpc.MSRPCNameArray)
+        assert isinstance(users, MSRPCNameArray)
         self.get_bytes()[4:-8] = array.array('B', users.rawData())
 
     def get_entries_num(self):
@@ -655,6 +825,14 @@ class DCERPCSamr:
     def __init__(self, dcerpc):
         self._dcerpc = dcerpc
 
+    def doRequest(self, request, noAnswer = 0, checkReturn = 1):
+        self._dcerpc.call(request.opnum, request)
+        if noAnswer:
+            return
+        else:
+            answer = self._dcerpc.recv()
+            return answer
+
     def connect(self):
         samrcon = SAMRConnectHeader()
         samrcon.set_server('*SMBSERVER')
@@ -689,9 +867,10 @@ class DCERPCSamr:
         retVal = SAMRRespOpenDomainHeader(data)
         return retVal
 
-    def enumusers(self,context_handle):
+    def enumusers(self,context_handle, resume_handle = 0):
         enumusers = SAMREnumDomainUsersHeader()
         enumusers.set_context_handle(context_handle)
+        enumusers.set_resume_handle(resume_handle)
         self._dcerpc.send(enumusers)
         data = self._dcerpc.recv()
         retVal = SAMRRespEnumDomainUsersHeader(data)
@@ -721,3 +900,42 @@ class DCERPCSamr:
         data = self._dcerpc.recv()
         retVal = SAMRRespCloseRequestHeader(data)
         return retVal
+
+    def EnumerateAliasesInDomain(self, context_handle):
+        enumAliases = SAMREnumerateAliasesInDomain()
+        enumAliases['ContextHandle'] = context_handle
+        ans = self.doRequest(enumAliases, checkReturn = 0)
+        packet = SAMREnumerateAliasesInDomainResponse(ans)
+        enum = MSRPCNameArray(packet['pEnumerationBuffer'])
+        return enum.elements()
+
+    def OpenAlias(self, context_handle, alias_id):
+        open_alias = SAMROpenAlias()
+        open_alias['ContextHandle'] = context_handle
+        open_alias['AliasId'] = alias_id
+        open_alias['AccessMask'] = 0x2000C
+        ans = self.doRequest(open_alias)
+        packet = SAMROpenAliasResponse(ans)
+        return packet
+
+    def GetMembersInAlias(self, context_handle):
+        alias_members = SAMRGetMembersInAlias()
+        alias_members['ContextHandle'] = context_handle
+        ans = self.doRequest(alias_members)
+        packet = SAMRGetMembersInAliasResponse(ans)
+        # Now parse the Aliases
+        if packet['Count'] > 0:
+           # Skipping the pointer data
+           data = packet['pEnumerationBuffer'][8:]
+           # Skipping the referent ID for each entry
+           data = data[4*packet['Count']:]
+        entries = []
+        for i in range(packet['Count']):
+           # Skip the count ID
+           data = data[4:]
+           entry = SAMR_RPC_SID(data)
+           entries.append(entry)
+           data = data[len(entry):]
+        packet['EnumerationBuffer'] = entries 
+        return packet
+
